\\ Copyright 2014, 2015, 2016 Kevin Ryde
\\
\\ This file is free software; you can redistribute it and/or modify it
\\ under the terms of the GNU General Public License as published by the Free
\\ Software Foundation; either version 3, or (at your option) any later
\\ version.
\\
\\ This file is distributed in the hope that it will be useful, but
\\ WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
\\ or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
\\ for more details.
\\
\\ You should have received a copy of the GNU General Public License along
\\ with this file.  See the file COPYING.  If not, see
\\ <http://www.gnu.org/licenses/>.

print("GP version ",version())

verbose = 1
test_count = 0;
test_equal(got,want,name="")=
{ test_count++;
  \\ print("got  ",got);
  \\ print("want ",got);
  if(got!=want,
     if(name!="",print1(name,": "));
     print("oops\ngot =",got,"\nwant=",want);
     quit(1));
  if(name=="",print("ok"),print(name,"  ok"));
}

\\ Something fishy in gp 2.5.5 on quad==quad part of tests.
\\ Add-back checks in pol_partial_fractions() itself pass.

\\-------------------------------
\\ type names

test_equal(type(I),           "t_COMPLEX");
test_equal(type(quadgen(-3)), "t_QUAD");
test_equal(type(1/'x),        "t_RFRAC");

\\-------------------------------
\\ pol_partial_fractions__find_factor -- internal

test_equal(pol_partial_fractions__find_factor(1-x), 0, "none");
test_equal(pol_partial_fractions__find_factor(I),   I, "complex");
test_equal(pol_partial_fractions__find_factor(quadgen(-3)*x^5), quadgen(-3), "quad");

\\-------------------------------
\\ pol_partial_fractions_merge

test_equal(pol_partial_fractions_merge([1/(1-x), 1/(1-x)]), \
           [2/(1-x)], \
           "merge 2->1");
test_equal(pol_partial_fractions_merge([1/(1-x), -1/(1-x)]), \
           [], \
           "merge 2->zero");
test_equal(pol_partial_fractions_merge([1/(1-x), 1/(1+x)]), \
           [1/(1-x), 1/(1+x)], \
           "merge");

test_equal(pol_partial_fractions_merge([1/(1-x+x^2), 1/(1+x)]), \
           [1/(1-x+x^2), 1/(1+x)], \
           "merge");

test_equal(pol_partial_fractions_merge([1/(1-x), 1/(1-x+x^2), -1/(1-x)]), \
           [1/(1-x+x^2)], \
           "merge");
{
  my(v = [1/(1-x), 1/(1-x)]);
  test_equal(pol_partial_fractions_merge(v), [2/(1-x)], "merge");
  test_equal(v, [1/(1-x), 1/(1-x)], "merge no modify");
}

\\-------------------------------
\\ pol_partial_fractions()

{
  my (f = 'y*'z/(1-x) );
  my (v = pol_partial_fractions(f));
  test_equal(v, [ 'y*'z/(1-'x) ], "yz ");
  test_equal(vecsum(v), f, "yz");
}           

\\ from the web page
{
  my (f = 2/(x^2-1) );
  my (v = pol_partial_fractions(f));
  test_equal(v, [ 1/(x-1),  -1/(x+1) ], "web");
  test_equal(vecsum(v), f, "web");
}           

\\ from the docs
{
  my(f=1/(x^2 - 1));
  my (v = pol_partial_fractions(f));
  test_equal(v, [ 1/(2*x - 2),  -1/(2*x + 2) ], "docs");
  test_equal(vecsum(v), f, "docs");
}           
{
  my (v = pol_partial_fractions( (1 + 2*I*x)/(x^2 + 1) ));
  test_equal(v, [I/(2*x - 2*I), 3*I/(2*x + 2*I)], "docs");
}           
{
  my(w = quadgen(-3));    \\ sixth root of unity
  my(v = pol_partial_fractions( 1/(1+x^3), w ));
  if(!(version()[1]==2 && version()[2]==5),
     test_equal(v, [1/(3*x + 3), -w/(3*x - 3*w), (-1 + w)/(3*x + (-3 + 3*w))], \
                "docs quadgen(-3)");
    );
}
{
  my(v = pol_partial_fractions( 1/(x^4+10*x^3+35*x^2+50*x+24),  [x+4] ));
  test_equal(v, [ -1/(6*x + 24),
                 (x^2 + 2*x + 3)/(6*x^3 + 36*x^2 + 66*x + 36)],
            "docs");
  test_equal(x^4+10*x^3+35*x^2+50*x+24, (x+1)*(x+2)*(x+3)*(x+4),
            "(x+1)*(x+2)*(x+3)*(x+4)");
}
{
  \\ factors column
  my(f = 2/(x^2-1));
  my(v = pol_partial_fractions(f, factor(denominator(f))[,1]));
  test_equal(v, [1/(x-1), -1/(x+1)], "factors column");
}
{
  my(v = pol_partial_fractions( x^6/(x^2-1) ));
  test_equal(v, [  x^4 + x^2 + 1,  1/(2*x - 2),  -1/(2*x + 2) ],
            "quotient term");
}
{
  my(v = pol_partial_fractions_merge([1/(x-1), x/(x-1), 1/(x+1)]));
  test_equal(v,[(x+1)/(x-1), 1/(x+1)]);
}

\\-----------------------------------------------------------------------------

test_equal(pol_partial_fractions(0), [], "zero");
test_equal(pol_partial_fractions(1), [1], "not a fraction");
test_equal(pol_partial_fractions(1+x+x^2), [1+x+x^2], "not a fraction");

test_equal(pol_partial_fractions(1/(x+1)), \
           [1/(x+1)], \
           "single");
test_equal(pol_partial_fractions(1/(x+1)^2), \
           [1/(x+1)^2]);
test_equal(pol_partial_fractions(1/(x+1)^3), \
           [1/(x+1)^3]);

test_equal(pol_partial_fractions(1/(x+1) + 1/(x-1)), \
           [1/(x-1), 1/(x+1)], \
           "x-1 and x+1");

test_dummy_var = 123;
{
  my(f = 1/('test_dummy_var+1) + 1/('test_dummy_var-1));
  test_equal(type(f), "t_RFRAC");
  test_equal(variable(f), 'test_dummy_var);
  test_equal(pol_partial_fractions(f),
             [1/('test_dummy_var-1), 1/('test_dummy_var+1)],
             "with test_dummy_var");
}

test_equal(pol_partial_fractions(1/(x+1) + 1/(x+1)^2), \
           [1/(x+1), 1/(x+1)^2], \
           "x+1 and (x+1)^2");

test_equal(pol_partial_fractions(x^6/( (1-x)*(1+x) ),  [1+x]), \
           [-x^4 - x^2 - 1, 1/(2*x + 2), 1/(-2*x + 2)]);


test_equal(pol_partial_fractions(1/(x+1)^2 + 1/(x+1)^3), \
           [1/(x+1)^2, 1/(x+1)^3], \
           "(x+1)^2 and (x+1)^3");

test_equal(pol_partial_fractions(1/(x+1) + 1/(x+1)^3, [(x+1)^2]), \
           [1/(x+1) + 1/(x+1)^3],  /* cannot break */ \
           "(x+1)^1 and (x+1)^3");

test_equal(pol_partial_fractions(1/(x+1) + 1/(x+1)^2 + 1/(x+1)^3, [(x+1)^2]), \
           [1/(x+1) + 1/(x+1)^2 + 1/(x+1)^3],  /* cannot break */ \
           "(x+1)^1 and (x+1)^2 and (x+1)^3");

{
for(k=1,10,
   my(v = vector(k,i, 1/(1-x)^i));
   my(f = vecsum(v));
   test_equal(pol_partial_fractions(f), v,
              Strprintf("powers k=%d",k)))
}
{
for(k=1,10,
   my(v = vector(k,i, (1-x)/(1-x-x^2)^i));
   my(f = vecsum(v));
   test_equal(pol_partial_fractions(f), v,
              Strprintf("powers k=%d",k)))
}

{
  my(gL_sans_first = (x)-> 2* (1 + x^2) / (1 - x - 2*x^3));
  my(b = 1 + I);
  my(v = [
          -1/2 + I,
          + 1/25 * (2 - 6*I)   / (1 + x),
          - 1/10 * (2 + I)     / (1 - conj(b)*x),
          + 1/50 * (61 - 23*I) / (1 - b*x),
          - 1/5  * (3 + I)     / (1 - b*x)^2,
          + 1/2 * gL_sans_first(b*x)
         ]);
  my(f = vecsum(v));
  test_equal(pol_partial_fractions(f), v, "gGLN")
}
{
  my(v = pol_partial_fractions( (1+x+x^2)/(x+1)^3 ));
  test_equal(v, [  1 / (x + 1),
                 -1 / (x^2 + 2*x + 1),              \\ (x+1)^2
                 1 / (x^3 + 3*x^2 + 3*x + 1) ],     \\ (x+1)^3
            "powers");
}
{
  my(v = pol_partial_fractions( 1/(x+1)^3 ));
  test_equal(v, [ 1/(x^3 + 3*x^2 + 3*x + 1) ],
             "f unchanged");
}
test_equal(pol_partial_fractions(1/(x^2+1)), \
           [1/(x^2+1)], \
           "x^2+1 over rationals");
test_equal(pol_partial_fractions(1/(x^2+1), I), \
           [-I/(2*x - 2*I), I/(2*x + 2*I)], \
           "x^2+1 over complex");
{
  if(!(version()[1]==2 && version()[2]==5),
     my(w = quadgen(-16)); \\ 2*I
     test_equal(w^2, -4);
     test_equal(w, 2.0*I);
     test_equal(pol_partial_fractions(1/(x^2+1), w), \
                [-w/(4*x - 2*w), w/(4*x + 2*w)], \
                "x^2+1 over quad 2*I");
     test_equal(-w/(4*x - 2*w) + w/(4*x + 2*w),
                1.0/(x^2+1));
    );
}

test_equal(pol_partial_fractions(1/(x^3+1), I), \
           [1/(3*x + 3), (-x + 2)/(3*x^2 - 3*x + 3)], \
           "x^3+1 over rationals");
{
  if(!(version()[1]==2 && version()[2]==5),
     my(w = quadgen(-3));  \\ sixth root of unity
     test_equal(pol_partial_fractions(1/(x^3+1), w),
                [1/(3*x + 3), -w/(3*x - 3*w), (-1 + w)/(3*x + (-3 + 3*w))],
                "x^3+1 over quad sixth root");
    );
}

\\-------------------------------
\\ when powers give gcd != 1

{
  my(f=1/(x+1) + 1/(x+1)^2 + 1/(x+1)^3);
  my(n=numerator(f));
  my(d=denominator(f));
  test_equal(n, x^2 + 3*x + 3, "n");
  test_equal(d, x^3 + 3*x^2 + 3*x + 1, "d");
  my(X=(x+1)^2);
  my(Y=d/X);
  my(a,b,g);
  my(v = gcdext(Y, X),     \\ find a*y + b*x = g
     a = v[1],
     b = v[2],
     g = v[3]);
  test_equal(g,x+1,"g");
  my(v = divrem(n,g),
     q = v[1],
     r = v[2]);
  test_equal(q,x + 2,"divrem(n,g) q");
  test_equal(r,1,"divrem(n,g) r");
  test_equal((x+2)*g + 1, n, "q*g+r==n");
}
{
  my(f=1/(x+1) + 1/(x+1)^2);
  my(n=numerator(f));
  my(d=denominator(f));
  test_equal(n, x + 2, "n");
  test_equal(d, x^2 + 2*x + 1, "d");
  my(X=(x+1)^2);
  my(Y=d/X);
  my(a,b,g);
  my(v = gcdext(Y, X),     \\ find a*y + b*x = g
     a = v[1],
     b = v[2],
     g = v[3]);
  test_equal(g,1,"g");
  my(q,r);
  my(v = divrem(n,g),
     q = v[1],
     r = v[2]);
  test_equal(q,x + 2,"divrem(n,g) q");
  test_equal(r,0,"divrem(n,g) r");
  test_equal((x+2)*g + 0, n, "q*g+r==n");
}

\\-------------------------------

\\ no globals changed
test_equal(a, 'a)
test_equal(b, 'b)
test_equal(e, 'e)
test_equal(f, 'f)
test_equal(j, 'j)
test_equal(n, 'n)
test_equal(p, 'p)
test_equal(q, 'q)
test_equal(r, 'r)
test_equal(s, 's)
test_equal(v, 'v)
test_equal(w, 'w)
test_equal(x, 'x)
test_equal(y, 'y)
test_equal(ret, 'ret)
test_equal(dens, 'dens)
test_equal(pos, 'pos)

\\-------------------------------
show_func_type(name) =
{
  my(f = eval(name));
  print(name," "type(f)," ",
        if(Str(f)==name, "gp2c compiled", "gp source"));
}
show_func_type("pol_partial_fractions");
show_func_type("pol_partial_fractions_merge");

\\-------------------------------
print(test_count," tests ok");




\\ Secret idea to take a vector for f.
\\
\\   if(type(f) == "t_VEC",
\\      pol_partial_fractions_merge
\\        (concat(apply(one -> pol_partial_fractions__one(one,dens), f))),
\\      pol_partial_fractions__one(f,dens));
\\ }
\\ pol_partial_fractions__one(f, dens) =
\\ {
