\\ Copyright 2014 Kevin Ryde
\\
\\ This file is free software; you can redistribute it and/or modify it
\\ under the terms of the GNU General Public License as published by the Free
\\ Software Foundation; either version 3, or (at your option) any later
\\ version.
\\
\\ This file is distributed in the hope that it will be useful, but
\\ WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
\\ or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
\\ for more details.
\\
\\ You should have received a copy of the GNU General Public License along
\\ with this file.  See the file COPYING.  If not, see
\\ <http://www.gnu.org/licenses/>.


pol_partial_fractions__find_quad(f) =
{
  f = concat(Vec(numerator(f)), Vec(denominator(f)));
  for(i=1,#f, if(type(f[i]) == "t_QUAD", return(f[i])));
  I;
}

pol_partial_fractions(f, dens=0) =
{
  \\ default denominators from factor()
  if(dens==0,
     my(m = factor(denominator(f)
                   * pol_partial_fractions__find_quad(f)));
     print("factor matrix ",m);
     dens = List([]);
     for(i=1,matsize(m)[1],
        for(j=1,m[i,2],
           listput(dens, m[i,1])));
     print("factor dens ",dens);
    );

  \\ dens argument check
  for(i=1,#dens,
     my(den = dens[i]);
     print("i=",i," den=",den);
     if(type(den) != "t_POL", error("denominator not a polynomial"));
     if(poldegree(den) < 1, error("denominator is a constant")));

  my(d = prod(i=1,#dens, dens[i]),
     total_unknowns = sum(i=1,#dens, poldegree(dens[i])));

  \\ extra denominator if only some dens given
  my(num=f*d);
  if(poldegree(denominator(num)) > 0,
     my(extra_den = denominator(num));
     print("extra den ",extra_den);
     d *= extra_den;
     total_unknowns += poldegree(extra_den);
     dens = concat([extra_den],dens);
     num = numerator(num));

  print("total denominator d=",d);
  print("total_unknowns ",total_unknowns);
  print("num=",num);

  my(ret = List([]));

  \\ quotient num/d if numerator has higher degree than d
  if (poldegree(num) >= total_unknowns,
      my(q);
      [q,num] = divrem(num,d);
      if(q, listput(ret,q));
      print("num higher degree q=",q," remaining num=",num);
     );

  my(m = matrix(total_unknowns, total_unknowns),
     u = 1);
  for(i=1,#dens,
     my(p = prod(j=1,#dens,
                 if(j<=i && dens[j]==dens[i],
                    1,
                    dens[j])));

     \\ (A+B*x)*p = A*p + x*B*p
     my(unknowns = poldegree(dens[i]),
        psize = poldegree(p)+1);
     print("unknowns=",unknowns," psize=",psize," p=",p);
     for(r=1, unknowns,
        print("column ",Col(Colrev(p,psize),total_unknowns));
        m[,u] =         Col(Colrev(p,psize),total_unknowns);
        u++; psize++));
  print("m=",m);
  print("det=",matdet(m)," kernel=",matker(m));

  print("num vector=",      Col(Colrev(num),total_unknowns));
  my(solution = matsolve(m, Col(Colrev(num),total_unknowns)));
  print("solution=",solution);
  
  u = 1;
  for(i=1,#dens,
     my(den = dens[i],
        unknowns = poldegree(den));

     print("den=",den);
     forstep(j=i-1, 1, -1,
            if(dens[j] == dens[i],
               print("ramified denominator");
               den *= dens[j],
               break));

     print("partial num = ",Polrev(solution[u .. u+unknowns-1]));
     my(n = Polrev(solution[u .. u+unknowns-1]));
     if(n, listput(ret,n/den));
     u += unknowns);

  ret = Vec(ret);
  
  \\ add as a check
  my(s = vecsum(ret));
  if (s == f,
      print("sum good"),
      print("oops different, got ",s," want ",f));

  ret;
}

